#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function
import os
import json
import shutil
import glob
import logging
from typing import List
from hashlib import blake2b
from signal import signal, SIGPIPE, SIGINT, SIG_DFL

signal(SIGPIPE, SIG_DFL)
signal(SIGINT, SIG_DFL)

SUFFIXES = ['.nsq', '.nin', '.nhr', '.nto', '.not', '.ndb', '.ntf']


def remove_intermediates(runid: str, intermediates: List[str], protect: bool) -> None:
    logger = logging.getLogger(__name__)
    if protect:
        msg = 'One or more intermediates of {} type need to be removed'
        logger.info(msg.format(','.join(intermediates)))
        msg = (
            'Delete the "protect" flag in the config.json file or restart '
            'your analysis with another runid to continue.'
        )
        logger.info(msg)
        end_program(1)

    genome_tmpfiles = [
        'blast_results.tsv',
        'keepsidx.json',
        'single_copy.json',
        'blast_filtered.tsv',
        'expected_filt.json',
    ]
    if 'genome' in intermediates:
        for deldir in ['unaligned', 'aligned']:
            if os.path.exists(deldir):
                shutil.rmtree(deldir)
        for fullname in [os.path.join('.autoMLSA', fn) for fn in genome_tmpfiles]:
            if os.path.exists(fullname):
                os.remove(fullname)
    for delfile in glob.iglob(runid + '.nex*'):
        os.remove(delfile)


def end_program(code: int) -> None:
    """
    Program message including success or failure of the program.
    """
    logger = logging.getLogger(__name__)
    if code == 1:
        msg = 'Program was stopped at an intermediate stage.'
        logger.info(msg)
    elif code == 0:
        msg = 'Program was a success! Congratulations!'
        logger.info(msg)
    else:
        msg = 'Program exiting with code ({}) indicating failure.'
        logger.info(msg.format(code))
        msg = 'Check error messages to resolve the problem.'
        logger.info(msg)
    exit(code)


def check_if_fasta(fa):
    """
    Checks first line of file for '>' to determine if FASTA, returns bool
    """
    for suffix in SUFFIXES:
        if suffix in fa:
            return False
    with open(fa, 'r') as fah:
        try:
            test = fah.readline().startswith('>')
        except UnicodeDecodeError:
            return False
        else:
            if test:
                return True
            else:
                return False


def json_writer(fn, x):
    """
    Writes object to file as json format.
    """
    with open(fn, 'w') as fh:
        json.dump(x, fh, indent=4)
        fh.write('\n')


def sanitize_path(s):
    """
    Turns string into filename without symbols
    """
    s = s.replace(' ', '_')
    filename = ''.join(x for x in s if (x.isalnum() or x in '._-'))
    return filename


def generate_hash(s):
    """
    Generates blake2b hash for sequence
    input  - nucleotide sequence
    return - blake2b hash digest
    """
    seqhash = blake2b(s.encode(), digest_size=16)
    return seqhash.hexdigest()


def checkpoint_reached(stage):
    logger = logging.getLogger(__name__)
    msg = 'Checkpoint reached {}. Stopping...'
    logger.info(msg.format(stage))
    end_program(1)


def checkpoint_tracker(fn_name):
    checkpath = os.path.join('.autoMLSA', 'checkpoint', fn_name)
    if not os.path.exists(checkpath):
        open(os.path.join(checkpath), 'w').close()


def exit_successfully(rundir: str, treefile: str) -> None:
    """Temporary command

    input  - rundir and treefile to print to log
    return - None
    """
    logger = logging.getLogger(__name__)

    msg = 'Your treefile is ready: {}/{}'
    logger.info(msg.format(rundir, treefile))
    end_program(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function
import os.path
import subprocess
import logging
import platform
import urllib.request
import shutil
from pathlib import Path
from hashlib import md5
from typing import Dict
from .helper_functions import end_program
from signal import signal, SIGPIPE, SIGINT, SIG_DFL

signal(SIGPIPE, SIG_DFL)
signal(SIGINT, SIG_DFL)

BLASTVER = '2.10.1'
MAFFTVER = '7.471'
IQTREEVER = '2.1.1'
EXTERNALDIR = os.path.join(os.path.expanduser('~'), '.local', 'external')


def get_external_path(external: str = '') -> str:
    """
    Gets path to external directory
    input  - None
    return - Path to external dir
    """

    if not external:
        external = EXTERNALDIR

    if not os.path.exists(external):
        os.makedirs(external)
    return external


def check_program(exe: str, program_name: str, version_flag: str) -> bool:
    logger = logging.getLogger(__name__)
    try:
        ver: str = subprocess.run(
            [exe, version_flag],
            check=True,
            stdout=subprocess.PIPE,
            text=True,
            stderr=subprocess.STDOUT,
        ).stdout
    except FileNotFoundError:
        msg = 'Unable to find {} executable in path or in provided dir ({})'
        logger.error(msg.format(program_name, exe))
        success = False
    except PermissionError:
        msg = (
            'Unable to run {} executable in ({}). Make sure it is installed properly '
            'and has the correct permissions to run.'
        )
        logger.error(msg.format(program_name, exe))
        success = False
    else:
        logger.debug('{} found: {}'.format(program_name, ver))
        success = True
    return success


def validate_requirements(external: str = '') -> Dict[str, str]:
    """
    Identifies installed software and sets paths for running

    input  - logger
    output - exes dict with paths to executables
    """
    logger = logging.getLogger(__name__)

    # Set local path to blast executable DIR
    BLASTPATH: str = ''
    externalpath = get_external_path(external)
    blastcheck = os.path.join(externalpath, 'ncbi-blast-{}+'.format(BLASTVER), 'bin')
    if os.path.exists(blastcheck):
        logger.debug('BLAST found in external dir.')
        BLASTPATH = blastcheck

    if platform.system() == 'Linux':
        mafftsuffix = 'linux64'
        iqtreesuffix = os.path.join('{}-Linux'.format(IQTREEVER), 'bin')
    elif platform.system() == 'Darwin':
        mafftsuffix = 'mac'
        iqtreesuffix = os.path.join('{}-MacOSX'.format(IQTREEVER), 'bin')
    elif platform.system() == 'Windows':
        mafftsuffix = ''
        iqtreesuffix = os.path.join('{}-Windows'.format(IQTREEVER), 'bin')

    mafftcheck = os.path.join(externalpath, 'mafft-{}'.format(mafftsuffix))
    iqtreecheck = os.path.join(externalpath, 'iqtree-{}'.format(iqtreesuffix))

    # Expects in $PATH, can hard-code to full path here
    tblastn: str = 'tblastn'
    blastn: str = 'blastn'
    makeblastdb: str = 'makeblastdb'
    mafft: str = 'mafft'
    if os.path.exists(mafftcheck):
        logger.debug('MAFFT found in external dir.')
        mafft = os.path.join(mafftcheck, 'mafft.bat')
    iqtree: str = 'iqtree2'
    if os.path.exists(iqtreecheck):
        logger.debug('iqtree found in external dir.')
        iqtree = os.path.join(iqtreecheck, 'iqtree2')

    if os.path.exists(os.path.join(BLASTPATH, tblastn)):
        tblastn = os.path.join(BLASTPATH, tblastn)
    if os.path.exists(os.path.join(BLASTPATH, blastn)):
        blastn = os.path.join(BLASTPATH, blastn)
    if os.path.exists(os.path.join(BLASTPATH, makeblastdb)):
        makeblastdb = os.path.join(BLASTPATH, makeblastdb)

    logger.debug('Checking tblastn: {}'.format(tblastn))
    if not check_program(tblastn, 'tblastn', '-version'):
        msg = 'Please add tblastn to your $PATH or supply the path as ' '--external'
        logger.error(msg)
        end_program(78)

    logger.debug('Checking blastn: {}'.format(blastn))
    if not check_program(blastn, 'blastn', '-version'):
        msg = 'Please add blastn to your $PATH or supply the path as ' '--external'
        logger.error(msg)
        end_program(78)

    logger.debug('Checking makeblastdb: {}'.format(makeblastdb))
    if not check_program(makeblastdb, 'makeblastdb', '-version'):
        msg = 'Please add makeblastdb to your $PATH or supply the path as ' '--external'
        logger.error(msg)
        end_program(78)

    # MAFFT testing
    logger.debug('Checking mafft: {}'.format(mafft))
    if not check_program(mafft, 'mafft', '--version'):
        msg = 'Please add mafft to your $PATH or supply the path as ' '--external'
        logger.error(msg)
        end_program(78)

    # iqtree testing
    logger.debug('Checking iqtree: {}'.format(iqtree))
    if not check_program(iqtree, 'iqtree', '--version'):
        msg = 'Please add iqtree to your $PATH or supply the path as ' '--external'
        logger.error(msg)
        end_program(78)

    exes: Dict[str, str] = {}
    exes['tblastn'] = tblastn
    exes['blastn'] = blastn
    exes['makeblastdb'] = makeblastdb
    exes['mafft'] = mafft
    exes['iqtree'] = iqtree
    return exes


def download_file(out_file: str, url: str, program: str) -> None:
    """
    Downloads given file from url
    """
    logger = logging.getLogger(__name__)
    logger.info('Downloading {} from {}.'.format(program, url))
    if not os.path.exists(out_file) or os.path.getsize(out_file) == 0:
        with urllib.request.urlopen(url) as response, open(out_file, 'wb') as tarfh:
            shutil.copyfileobj(response, tarfh)
    else:
        logger.info('{} already downloaded.'.format(program))


def install_blast(external: str = '') -> None:
    """
    Installs BLAST executables to external dir
    """
    logger = logging.getLogger(__name__)
    externalpath = get_external_path(external)
    base_url = 'https://ftp.ncbi.nlm.nih.gov/blast/executables/blast+/{}'.format(
        BLASTVER
    )
    base_file = 'ncbi-blast-{}+-x64-'.format(BLASTVER)
    if platform.system() == 'Linux':
        base_out_file = base_file + 'linux.tar.gz'
    elif platform.system() == 'Darwin':
        base_out_file = base_file + 'macosx.tar.gz'
    elif platform.system() == 'Windows':
        base_out_file = base_file + 'win64.tar.gz'
    out_file = os.path.join(externalpath, base_out_file)
    blast_tar_url = '/'.join([base_url, base_out_file])
    md5_url = '/'.join([base_url, base_out_file + '.md5'])
    download_file(out_file, blast_tar_url, 'BLAST')

    blast_md5_digest = md5(Path(out_file).read_bytes()).hexdigest()
    logger.debug('calculated md5 digest: {}'.format(blast_md5_digest))

    download_file(out_file + '.md5', md5_url, 'BLAST md5')
    with open(out_file + '.md5', 'r') as md5fh:
        md5_hash = md5fh.readline().split(' ', 1)[0]
    logger.debug('given md5 digest: {}'.format(md5_hash))

    if blast_md5_digest != md5_hash:
        logger.info('BLAST download was unsuccessful. Delete file and try again.')
        end_program(-1)
    tblastn = os.path.join(
        externalpath, 'ncbi-blast-{}+'.format(BLASTVER), 'bin', 'tblastn'
    )
    if not os.path.exists(tblastn):
        logger.info('Unpacking BLAST tar.gz file.')
        shutil.unpack_archive(out_file, externalpath, 'gztar')
    logger.info('Checking tblastn version.')
    try:
        tblastn_ver: bytes = subprocess.check_output(
            [tblastn, '-version'], stderr=subprocess.STDOUT
        )
    except FileNotFoundError:
        msg = 'Unable to find tblastn executable in path or in provided dir ({})'
        logger.error(msg.format(tblastn))
        msg = 'Please add tblastn to your $PATH or supply a BLASTPATH in {}'
        logger.error(msg.format(__file__))
        end_program(78)
    else:
        logger.info('tblastn found: {}'.format(tblastn_ver.strip().decode()))
        logger.info(
            'Add {} to your $PATH to have access to this program outside of automlsa2'.format(
                os.path.dirname(tblastn)
            )
        )


def install_mafft(external: str = '') -> None:
    logger = logging.getLogger(__name__)

    base_url = 'https://mafft.cbrc.jp/alignment/software'
    base_file = 'mafft-{}'.format(MAFFTVER)
    if platform.system() == 'Linux':
        base_out_file = base_file + '-linux.tgz'
        md5_hash = '641cf15f90fc6634f9b2b2c7cf67e27d'
    elif platform.system() == 'Darwin':
        base_out_file = base_file + '-mac.zip'
        md5_hash = 'f57cd93ea546e47fbad975aecb58d303'
    elif platform.system() == 'Windows':
        logger.error('Windows is not currently supported for MAFFT install.')
        logger.error(
            'Look at install options here: '
            'https://mafft.cbrc.jp/alignment/software/windows.html'
        )
        return

    externalpath = get_external_path(external)
    out_file = os.path.join(externalpath, base_out_file)
    tar_url = '/'.join([base_url, base_out_file])
    download_file(out_file, tar_url, 'MAFFT')

    mafft_md5_digest = md5(Path(out_file).read_bytes()).hexdigest()
    logger.debug('calculated md5 digest: {}'.format(mafft_md5_digest))
    logger.debug('given md5 digest: {}'.format(md5_hash))
    if mafft_md5_digest != md5_hash:
        logger.info('MAFFT download was unsuccessful. Delete file and try again.')
        exit(-1)
    if platform.system() == 'Linux':
        mafft = os.path.join(externalpath, 'mafft-linux64', 'mafft.bat')
        if not os.path.exists(mafft):
            logger.info('Unpacking MAFFT tar.gz file.')
            shutil.unpack_archive(out_file, externalpath, 'gztar')
    elif platform.system() == 'Darwin':
        mafft = os.path.join(externalpath, 'mafft-mac', 'mafft.bat')
        if not os.path.exists(mafft):
            logger.info('Unpacking MAFFT zip file.')
            shutil.unpack_archive(out_file, externalpath, 'zip')
        logger.debug('Changing permissions.')
        for root, _, files in os.walk(os.path.join(externalpath, 'mafft-mac')):
            for exe in files:
                os.chmod(os.path.join(root, exe), 0o755)
    logger.info('Checking mafft version.')
    try:
        mafft_ver: bytes = subprocess.check_output(
            [mafft, '--version'], stderr=subprocess.STDOUT
        )
    except FileNotFoundError:
        msg = 'Unable to find mafft executable in path or in provided dir ({})'
        logger.error(msg.format(mafft))
        msg = 'Please add mafft to your $PATH or supply a BLASTPATH in {}'
        logger.error(msg.format(__file__))
        end_program(78)
    else:
        logger.info('mafft found: {}'.format(mafft_ver.strip().decode()))
        logger.info(
            'Add {} to your $PATH to have access to this program outside of automlsa2'.format(
                os.path.dirname(mafft)
            )
        )


def install_iqtree(external: str = '') -> None:
    logger = logging.getLogger(__name__)

    base_url = 'https://github.com/iqtree/iqtree2/releases/download/v{}'.format(
        IQTREEVER
    )
    base_file = 'iqtree-{}'.format(IQTREEVER)
    if platform.system() == 'Linux':
        base_out_file = base_file + '-Linux.tar.gz'
        md5_hash = '2e2dfeeb2a1e9e123f542b200f18340d'
    elif platform.system() == 'Darwin':
        base_out_file = base_file + '-MacOSX.zip'
        md5_hash = '45314fba262f200c32c37deceefbf5b1'
    elif platform.system() == 'Windows':
        base_out_file = base_file + '-Windows.zip'
        md5_hash = 'f1a25e89a5c31437125ae0c666d29ee8'

    externalpath = get_external_path(external)
    out_file = os.path.join(externalpath, base_out_file)
    tar_url = '/'.join([base_url, base_out_file])
    download_file(out_file, tar_url, 'iqtree')

    iqtree_md5_digest = md5(Path(out_file).read_bytes()).hexdigest()
    logger.debug('calculated md5 digest: {}'.format(iqtree_md5_digest))
    logger.debug('given md5 digest: {}'.format(md5_hash))
    if iqtree_md5_digest != md5_hash:
        logger.info('iqtree2 download was unsuccessful. Delete file and try again.')
        exit(-1)
    if platform.system() == 'Linux':
        iqtree = os.path.join(
            externalpath, 'iqtree-{}-Linux'.format(IQTREEVER), 'bin', 'iqtree2'
        )
        if not os.path.exists(iqtree):
            logger.info('Unpacking iqtree tar.gz file.')
            shutil.unpack_archive(out_file, externalpath, 'gztar')
    elif platform.system() == 'Darwin':
        iqtree = os.path.join(
            externalpath, 'iqtree-{}-MacOSX'.format(IQTREEVER), 'bin', 'iqtree2'
        )
        if not os.path.exists(iqtree):
            logger.info('Unpacking iqtree zip file.')
            shutil.unpack_archive(out_file, externalpath, 'zip')
        os.chmod(iqtree, 0o755)
    elif platform.system() == 'Windows':
        iqtree = os.path.join(
            externalpath, 'iqtree-{}-Windows'.format(IQTREEVER), 'bin', 'iqtree2.exe'
        )
        if not os.path.exists(iqtree):
            logger.info('Unpacking iqtree zip file.')
            shutil.unpack_archive(out_file, externalpath, 'zip')
    logger.info('Checking iqtree version.')
    try:
        iqtree_ver: bytes = subprocess.check_output(
            [iqtree, '--version'], stderr=subprocess.STDOUT
        )
    except FileNotFoundError:
        msg = 'Unable to find iqtree2 executable in path or in provided dir ({})'
        logger.error(msg.format(iqtree))
        msg = 'Please add iqtree2 to your $PATH or supply a full iqtree path in {}'
        logger.error(msg.format(__file__))
        end_program(78)
    else:
        logger.info('iqtree2 found: {}'.format(iqtree_ver.strip().decode()))
        logger.info(
            'Add {} to your $PATH to have access to this program outside of automlsa2'.format(
                os.path.dirname(iqtree)
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import glob
import json
import argparse
import subprocess
import pprint
from typing import List, Dict, Any
from .helper_functions import (
    check_if_fasta,
    end_program,
    checkpoint_reached,
    json_writer,
    remove_intermediates,
)


def get_fasta_list(dirpath: str, rundir: str) -> List[str]:
    """
    Finds FASTA files from directory
    """
    logger = logging.getLogger(__name__)
    if not os.path.exists(dirpath):
        fastadir = os.path.join(rundir, 'fasta')
        if os.path.exists(fastadir):
            msg = 'Original genome directory ({}) does not exist.'
            logger.warning(msg.format(dirpath))
            msg = 'Attempting to run from already renamed FASTA files.'
            logger.warning(msg)
            dirpath = fastadir
    fasta_list: List[str] = []
    logger.debug('Checking for FASTA format files in {}'.format(dirpath))
    for fa in glob.iglob(os.path.join(dirpath, '*')):
        # sys.stderr.write('Checking if {} is FASTA.\n'.format(fa))
        if os.path.isdir(fa):
            continue
        if check_if_fasta(fa):
            # sys.stderr.write('{} is FASTA.\n'.format(fa))
            fasta_list.append(os.path.abspath(fa))
        else:
            msg = '{} does not appear to be FASTA file, skipping.'
            logger.debug(msg.format(os.path.basename(fa)))
    if fasta_list:
        msg = 'Found FASTA files in {}'
        logger.info(msg.format(dirpath))
    return fasta_list


def read_config(configfile: str) -> Dict[str, Any]:
    """Finds settings in config file (json format). path/rundir/config.json

    input  - rundir path, logger
    return - dict with configuration settings
    """
    logger = logging.getLogger(__name__)

    logger.info('Reading from configuration file: {}'.format(configfile))

    with open(configfile, 'r') as cf:
        config: Dict[str, Any] = json.load(cf)
    return config


def validate_arguments(
    args: argparse.Namespace, config: dict, checkpoint: bool
) -> argparse.Namespace:
    """Checks for executables as well as makes sure files exist.

    input  - args from argparse; config from config file
    return - validated args from argparse
    """
    # config dict is from file
    # Determine if defaults need to be set
    # e = 1e-5
    # m = 500
    # c = 50
    # p = tblastn
    # t = 1

    logger = logging.getLogger(__name__)
    logger.debug('Validating & reconciling arguments.')
    error: bool = False

    autoMLSA = os.path.join(args.rundir, '.autoMLSA')
    checkpoint_dir = os.path.join(autoMLSA, 'checkpoint')

    if not os.path.exists(autoMLSA):
        os.mkdir(autoMLSA)

    if not os.path.exists(checkpoint_dir):
        os.mkdir(checkpoint_dir)

    # Reconcile config file and command line options
    changed: bool = False
    if args.evalue is None:
        if 'evalue' in config.keys() and config['evalue'] is not None:
            args.evalue = float(config['evalue'])
        else:
            args.evalue = 1e-5
    else:
        try:
            check = args.evalue != float(config['evalue'])
        except KeyError:
            pass
        else:
            if check:
                changed = True
    if args.coverage is None:
        if 'coverage' in config.keys() and config['coverage'] is not None:
            args.coverage = int(config['coverage'])
        else:
            args.coverage = 50
    else:
        try:
            check = args.coverage != int(config['coverage'])
        except KeyError:
            pass
        else:
            if check:
                changed = True
    if args.identity is None:
        if 'identity' in config.keys() and config['identity'] is not None:
            args.identity = int(config['identity'])
        else:
            args.identity = 30
    else:
        try:
            check = args.identity != int(config['identity'])
        except KeyError:
            pass
        else:
            if check:
                changed = True
    if args.allow_missing is None:
        if 'allow_missing' in config.keys() and config['allow_missing'] is not None:
            args.allow_missing = int(config['allow_missing'])
        else:
            args.allow_missing = 0
    else:
        try:
            check = args.allow_missing != int(config['allow_missing'])
        except KeyError:
            pass
        else:
            if check:
                changed = True
    if not args.missing_check:
        if 'missing_check' in config.keys() and config['missing_check']:
            args.allow_missing = True
        else:
            args.missing_check = False
    if args.program is None:
        if 'program' in config.keys():
            if config['program'] in ('tblastn', 'blastn'):
                args.program = config['program']
            else:
                msg = 'Program specified in config file "{}" is not valid.'
                logger.error(msg.format(config['program']))
                msg = 'Please give either "tblastn" or "blastn" and try again.'
                logger.error(msg)
                end_program(78)
        else:
            args.program = 'tblastn'
    else:
        try:
            check = args.program != config['program']
        except KeyError:
            pass
        else:
            if check:
                changed = True
    if args.threads is None:
        if 'threads' in config.keys() and config['threads'] is not None:
            args.threads = int(config['threads'])
        else:
            args.threads = 1
    if args.outgroup is None:
        if 'outgroup' in config.keys() and config['outgroup'] is not None:
            args.outgroup = config['outgroup']
        else:
            args.outgroup = ''
    else:
        try:
            check = args.outgroup != config['outgroup']
        except KeyError:
            pass
        else:
            if check:
                pass

    if args.iqtree is None:
        if 'iqtree' in config.keys() and config['iqtree'] is not None:
            args.iqtree = config['iqtree']
        else:
            args.iqtree = (
                '-m MFP -B 1000 -alrt 1000 --msub nuclear --merge ' 'rclusterf'
            )
    else:
        try:
            check = args.iqtree != config['iqtree']
        except KeyError:
            pass
        else:
            if check:
                changed = True

    if args.mafft is None:
        if 'mafft' in config.keys() and config['mafft'] is not None:
            args.mafft = config['mafft']
        else:
            args.mafft = '--localpair --maxiterate 1000'
    else:
        try:
            check = args.mafft != config['mafft']
        except KeyError:
            pass
        else:
            if check:
                changed = True

    if changed:
        logger.debug('One or more settings was changed from config.')
        remove_intermediates(args.runid, ['genome'], args.protect)

    if args.checkpoint is None:
        if 'checkpoint' in config.keys() and config['checkpoint'] is not None:
            args.checkpoint = config['checkpoint']
        else:
            args.checkpoint = 'none'
        if args.checkpoint == 'validate':
            checkpoint = True
    if args.protect is None:
        if 'protect' in config.keys() and config['protect'] is not None:
            args.protect = config['protect']
        else:
            args.protect = False

    if args.evalue > 10:
        msg = 'Specified evalue "{}" is greater than 10.'
        logger.error(msg.format(args.evalue))
        msg = 'Please specify an evalue < 10 and try again.'
        logger.error(msg)
        end_program(78)
    if args.coverage not in range(0, 100):
        msg = 'Coverage value is not between 0 and 100.'
        logger.error(msg.format(args.coverage))
        msg = 'Please specify a coverage between 0 and 100 and try again.'
        logger.error(msg)
        end_program(78)

    cmd = 'lscpu | grep -G "^CPU(s):" | grep -o -E "[0-9]+"'
    try:
        maxthreads = (
            subprocess.check_output(cmd, shell=True, stderr=subprocess.STDOUT)
            .decode()
            .strip()
        )
    except subprocess.CalledProcessError:
        msg = 'Unable to check number of available threads.'
        logger.warning(msg)
        msg = 'Make sure the number of specified threads is correct.'
        logger.warning(msg)
    else:
        logger.debug('Max threads found {}'.format(maxthreads))
        if args.threads > int(maxthreads):
            msg = 'Threads specified {} greater than number of available ' 'threads {}'
            logger.error(msg.format(args.threads, maxthreads))
            msg = 'Specify threads less than or equal to {} and try again.'
            logger.error(msg.format(maxthreads))
            end_program(78)

    # Set up lists for files, dirs, and queries
    if args.files is None:
        args.files = []
    if 'files' in config.keys() and config['files'] is not None:
        for f in config['files']:
            if f:
                args.files.append(f)

    if args.dir is None:
        args.dir = []
    if 'dir' in config.keys() and config['dir'] is not None:
        for d in config['dir']:
            if d:
                args.dir.append(d)

    if args.query is None:
        args.query = []
    if 'query' in config.keys() and config['query'] is not None:
        for q in config['query']:
            if q:
                args.query.append(q)
    for query in args.query.copy():
        if not check_if_fasta(query):
            msg = 'Specified query file {} does not appear to be FASTA file.'
            logger.error(msg.format(query))
            logger.debug('Removing {} from query list'.format(query))
            args.query.remove(query)
            error = True

    args.fasta = []
    # Combine lists
    for fastadir in args.dir:
        fastas = get_fasta_list(fastadir, args.rundir)
        args.fasta.extend(fastas)
    args.fasta.extend(args.files)

    # Remove duplicates
    args.fasta = list(dict.fromkeys(args.fasta))
    args.dir = list(dict.fromkeys(args.dir))
    args.query = list(dict.fromkeys(args.query))

    seen: Dict[str, str] = {}
    for fasta in args.fasta:
        base = os.path.basename(fasta)
        if base in seen:
            msg = 'Same genome name found in two locations'
            logger.error(msg)
            msg = '++++++++++++Offending files++++++++++++'
            logger.error(msg)
            logger.error('{}'.format(fasta))
            logger.error('{}'.format(seen[base]))
            msg = '+++++++++++++++++++++++++++++++++++++++'
            logger.error(msg)
            msg = 'Remove or rename one of these to continue.'
            logger.error(msg)
            end_program(65)
        else:
            seen[base] = fasta

    if 'dups' in config.keys() and config['dups']:
        args.dups = True

    if not args.fasta:
        msg = 'No valid FASTA files provided as --dir or --files.'
        logger.error(msg)
        error = True

    logger.debug('Validated arguments:')
    logger.debug('\n' + pprint.pformat(vars(args), indent=4))

    write_config(args)
    if error:
        end_program(-1)

    if checkpoint:
        checkpoint_reached('at validate arguments')

    return args


def write_config(args: argparse.Namespace) -> None:
    """Writes new config file, overwriting old if necessary

    input  - valiated & compiled arguments
    return - NULL
    """
    logger = logging.getLogger(__name__)
    logger.debug('Writing config file {}'.format(args.configfile))
    configdict = vars(args).copy()
    configdict.pop('config')
    configdict.pop('debug')
    configdict.pop('quiet')
    configdict.pop('configfile')
    configdict.pop('rundir')
    configdict.pop('checkpoint')
    json_writer(args.configfile, configdict)


def get_labels(rundir: str, fastas: List[str]) -> List[str]:
    """Determines the arbitrary labels per fasta file
    labels are index of list

    input - fasta files
    return - label list
    """
    logger = logging.getLogger(__name__)
    labelsf: str = os.path.join(rundir, '.autoMLSA', 'labels.json')
    bases: List[str] = [os.path.basename(x) for x in fastas]
    if os.path.exists(labelsf):
        with open(labelsf, 'r') as lf:
            labels: List[str] = json.load(lf)
        for base in bases:
            if base not in labels:
                logger.debug('Adding {} to list of labels'.format(base))
                labels.append(base)
    else:
        logger.debug('Generating list of labels {}'.format(labelsf))
        labels = bases
    json_writer(labelsf, labels)
    return labels



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function

# import sys
import argparse

# import csv
import os.path
import logging
import ete3  # type: ignore

# from collections import defaultdict
from signal import signal, SIGPIPE, SIGINT, SIG_DFL

signal(SIGPIPE, SIG_DFL)
signal(SIGINT, SIG_DFL)


def extant_file(x):
    """
    'Type' for argparse - checks that file exists but does not open.
    """
    if not os.path.exists(x):
        # Argparse uses the ArgumentTypeError to give a rejection message like:
        # error: argument input: x does not exist
        raise argparse.ArgumentTypeError("{0} does not exist".format(x))
    return x


def init_logger(args):
    logger = logging.getLogger(__name__)
    ch = logging.StreamHandler()
    if args.debug:
        logger.setLevel(logging.DEBUG)
        ch.setLevel(logging.DEBUG)
    elif args.verbose:
        logger.setLevel(logging.INFO)
        ch.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)
        ch.setLevel(logging.WARNING)
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


def run_argparse():
    parser = argparse.ArgumentParser(
        description='Tool to print tree images to file. Handles multiple '
        'bootstrap support values.'
    )
    parser.add_argument('treefile', help='Input newick format file.', type=extant_file)
    parser.add_argument(
        '--fsize', help='Default fontsize. [%(default)i]', type=int, default=10
    )
    parser.add_argument(
        '--ftype', help='Default fonttype. [%(default)s]', type=str, default='Calibri'
    )
    parser.add_argument(
        '--dpi',
        help='dots-per-inch value for output. [%(default)i]',
        type=int,
        default=300,
    )
    parser.add_argument(
        '--width',
        help='width (inches) value for output. [%(default)i]',
        type=int,
        default=5,
    )
    parser.add_argument(
        '--debug', help='Turn on debugging messages.', action='store_true'
    )
    parser.add_argument(
        '--verbose', help='Print verbose progress messages.', action='store_true'
    )
    args = parser.parse_args()
    args.logger = init_logger(args)
    return (args.logger, args.treefile, args.dpi, args.width, args.fsize, args.ftype)


def main():
    logger, treefile, dpi, width, fsize, ftype = run_argparse()
    # Testing
    t = ete3.Tree(treefile, format=1)
    ts = ete3.TreeStyle()
    ts.show_leaf_name = False
    ts.legend.add_face(
        ete3.TextFace('Legend', fsize=int(fsize * 0.9), ftype=ftype, penwidth=6),
        column=1,
    )
    ts.legend.add_face(ete3.TextFace(''), column=0)
    ts.legend.add_face(ete3.CircleFace(3, 'FireBrick'), column=0)
    ts.legend.add_face(
        ete3.TextFace('Well Supported', fsize=int(fsize * 0.8), ftype=ftype), column=1
    )
    ts.legend_position = 4
    # ts.scale = 120
    ts.branch_vertical_margin = 10
    # Convert to +/- support
    for i, node in t.traverse():
        if not node.is_leaf():
            if node.name:
                alrt, boot = node.name.split("/", maxsplit=1)
                node.add_feature('alrt', float(alrt))
                node.add_feature('bootstrap', int(boot))
            else:
                node.add_feature('alrt', float(0.0))
                node.add_feature('bootstrap', int(0))
            if node.alrt > 80 and node.bootstrap > 95:
                node.support = 100
            else:
                node.support = 0

    print('# Original')
    print(t)
    outgroup = t.get_midpoint_outgroup()
    t.set_outgroup(outgroup)
    print('# Midpoint Rooted')
    print(t)

    for i, node in enumerate(t.traverse()):
        nstyle = ete3.NodeStyle()
        if i == 0:
            nstyle['size'] = 0
        elif not node.is_leaf():
            nstyle['size'] = 6
            if node.support == 100:
                nstyle['fgcolor'] = 'FireBrick'
            else:
                nstyle['fgcolor'] = 'black'
        else:
            node.add_face(ete3.AttrFace('name', ftype=ftype, fsize=int(fsize)), 0)
            nstyle['size'] = 0
        node.set_style(nstyle)

    t.render(treefile + '.png', dpi=dpi, w=width, units='in', tree_style=ts)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import os
import json
import glob

# from tqdm import tqdm  # type: ignore
from rich.progress import track
from .helper_functions import (
    generate_hash,
    json_writer,
    remove_intermediates,
    end_program,
    sanitize_path,
)
from .blast_functions import make_blast_database
from Bio import SeqIO  # type: ignore
from collections import defaultdict
from typing import List, Dict, Any
from shutil import copy


def convert_fasta(
    rundir: str,
    fastas: List[str],
    labels: List[str],
    makeblastdb: str,
    runid: str,
    protect: bool,
) -> List[str]:
    """Converts fasta file with placeholder label names

    input - unformatted fasta files
    return - Set of formatted fasta files
    also generates BLAST dbs per fasta
    """
    logger = logging.getLogger(__name__)
    fastadir: str = os.path.join(rundir, 'fasta')
    new_fastas: List[str] = []
    renamef: str = os.path.join(rundir, '.autoMLSA', 'rename.json')
    deleted: bool = False
    if os.path.exists(renamef):
        with open(renamef, 'r') as rf:
            rename_info: Dict[Any, Any] = json.load(rf)
    else:
        rename_info = defaultdict(dict)
    if not os.path.exists(fastadir):
        os.mkdir(fastadir)

    for fasta in track(fastas, 'Working...'.rjust(19, ' ')):
        base = os.path.basename(fasta)
        label = labels.index(base)
        labelfastaf = os.path.join(fastadir, base)
        if base not in rename_info:
            try:
                rename_info[base]['index'] = label
            except KeyError:
                rename_info[base] = {}
                rename_info[base]['index'] = label
        else:
            if check_hash(fasta, base, rename_info[base]['hash']):
                deleted = True
        new_fastas.append(labelfastaf)
        if not os.path.exists(labelfastaf):
            msg = 'Writing renamed fasta for {}'
            logger.debug(msg.format(base))
            seq = ''
            with open(fasta, 'r') as ff, open(labelfastaf, 'w') as lff:
                for rec in SeqIO.parse(ff, 'fasta'):
                    lff.write('>{} {}\n'.format(label, rec.id))
                    lff.write('{}\n'.format(rec.seq))
                    seq += str(rec.seq)
            seqhash = generate_hash(seq)
            rename_info[base]['hash'] = seqhash
    logger.info('Generating and/or validating BLAST DBs.')
    for labelfastaf in track(new_fastas, 'makeblastdb...'.rjust(19, ' ')):
        make_blast_database(makeblastdb, labelfastaf)
    json_writer(renamef, rename_info)

    expected_fastas_fn: str = os.path.join(rundir, '.autoMLSA', 'expected_fastas.json')
    if os.path.exists(expected_fastas_fn):
        with open(expected_fastas_fn, 'r') as eff:
            expected_fastas: List[str] = json.load(eff)
    else:
        expected_fastas = []

    updated: bool = set(expected_fastas) != set(new_fastas)
    if updated or deleted:
        logger.debug('Found new genome files')
        remove_intermediates(runid, ['genome'], protect)
    else:
        logger.debug('Found no new genome files')
    json_writer(expected_fastas_fn, new_fastas)

    return new_fastas


def check_hash(fasta: str, base: str, seqhash: str) -> bool:
    """
    Checks hash, deletes files related to genome if necessary
    return - value if something is deleted
    """
    logger = logging.getLogger(__name__)
    deleted: bool = False
    with open(fasta, 'r') as ff:
        newseqhash = generate_hash(
            ''.join([str(rec.seq) for rec in SeqIO.parse(ff, 'fasta')])
        )
    if newseqhash != seqhash:
        deleted = True
        msg = 'Old genome file {} sequence has changed. Updating.'
        logger.info(msg.format(base))
        for fa in glob.iglob(os.path.join('fasta', base + '*')):
            os.remove(fa)
        for res in glob.iglob(os.path.join('blast', '*_' + base + '.tab')):
            os.remove(res)
    else:
        logger.debug('Genome file {} sequence has not changed.'.format(base))
    return deleted


def get_queries(
    runid: str, rundir: str, dups: bool, queries: List[str], protect: bool
) -> List[str]:
    """Converts query fasta file(s) into individual fastas for BLAST

    input - FASTA files with queries as individual entries
    return - individual fasta files
    """
    logger = logging.getLogger(__name__)
    querydir: str = os.path.join(rundir, 'queries')
    backups: str = os.path.join(rundir, '.autoMLSA', 'backups')

    new_queries: List[str] = []
    seen: Dict[str, Dict[str, str]] = {}
    hashes: Dict[str, str] = {}
    if not os.path.exists(querydir):
        os.mkdir(querydir)
    for query_file in queries:
        i = 1
        query_base: str = os.path.basename(query_file)
        query_backup: str = os.path.join(backups, query_base)
        if not os.path.exists(query_file):
            if os.path.exists(query_backup):
                msg = 'Using query backup file as {} is missing.'
                logger.warning(msg.format(query_file))
                query_file = query_backup
            else:
                msg = 'Query file {} seems to have been removed and/or lost.'
                logger.critical(msg.format(query_file))
                msg = 'No backup was found in the backups directory.'
                logger.critical(msg)
                msg = (
                    'Unable to continue without the queries. Either replace'
                    ' the file or start again from a new analysis.'
                )
                logger.critical(msg)
                end_program(66)
        with open(query_file, 'r') as qf:
            logger.debug('Reading query file ({})'.format(query_file))
            for rec in SeqIO.parse(qf, 'fasta'):
                safeid: str = sanitize_path(rec.id)
                seqhash: str = generate_hash(str(rec.seq))

                if seqhash in hashes and hashes[seqhash] == safeid:
                    # Skip duplicate header + seq
                    continue
                elif seqhash in hashes and hashes[seqhash] != safeid:
                    # Same seq but different names, throw error
                    mismatchid: str = hashes[seqhash]
                    msg = (
                        'Same sequence found in two query inputs with '
                        'different sequence IDs:'
                    )
                    logger.error(msg)
                    msg = '+++++++++++++++Offending queries+++++++++++++++'
                    logger.error(msg)
                    if query_file == seen[mismatchid]['path']:
                        msg = '{} - TWICE - header ID {} & {}'
                        logger.error(
                            msg.format(
                                os.path.basename(query_file),
                                seen[mismatchid]['unsafeid'],
                                rec.id,
                            )
                        )
                    else:
                        msg = '{} header ID {}'
                        logger.error(msg.format(os.path.basename(query_file), rec.id))
                        logger.error(
                            msg.format(
                                os.path.basename(seen[mismatchid]['path']),
                                seen[mismatchid]['unsafeid'],
                            )
                        )
                    msg = '+++++++++++++++++++++++++++++++++++++++++++++++'
                    logger.error(msg)
                    msg = (
                        'Check your sequences to make sure they aren\'t '
                        'misnamed, fix the problem, and try again.'
                    )
                    logger.error(msg)
                    end_program(65)
                else:
                    # Expected outcome; new id, new seq
                    hashes[seqhash] = safeid

                # Check for same id, allow if --dups flag, otherwise error
                if safeid in seen and not dups:
                    msg = 'Same query name ({}) found in two query inputs:'
                    logger.error(msg.format(safeid))
                    msg = '+++++++++++++++Offending queries+++++++++++++++'
                    logger.error(msg)
                    if query_file == seen[safeid]['path']:
                        msg = '{} - TWICE - sequences {} & {}'
                        logger.error(
                            msg.format(
                                os.path.basename(query_file), seen[safeid]['index'], i
                            )
                        )
                    else:
                        msg = '{} sequence {}'
                        logger.error(msg.format(os.path.basename(query_file), i))
                        logger.error(
                            msg.format(
                                os.path.basename(seen[safeid]['path']),
                                seen[safeid]['index'],
                            )
                        )
                    msg = '+++++++++++++++++++++++++++++++++++++++++++++++'
                    logger.error(msg)
                    msg = 'Remove or rename one of these to continue.'
                    logger.error(msg)
                    msg = (
                        'Alternatively, if this is intended, use the --dups'
                        ' flag to include both copies.'
                    )
                    logger.error(msg)
                    end_program(65)
                else:
                    if safeid in seen and dups:
                        msg = (
                            'Keeping additional query {} with duplicate ID '
                            'from file {}.'
                        )
                        logger.info(msg.format(safeid, os.path.basename(query_file)))
                    seen[safeid] = {}
                    seen[safeid]['path'] = query_file
                    seen[safeid]['index'] = str(i)
                    seen[safeid]['unsafeid'] = rec.id

                fn = os.path.join(querydir, safeid + '_' + seqhash + '.fas')
                new_queries.append(fn)
                if not os.path.exists(fn):
                    msg = 'Writing {} for seq {} in {}'
                    logger.debug(
                        msg.format(
                            os.path.basename(fn), i, os.path.basename(query_file)
                        )
                    )
                    with open(fn, 'w') as fh:
                        fh.write('>{}\n'.format(safeid))
                        fh.write('{}\n'.format(rec.seq))
                i += 1
        if query_file != query_backup:
            copy(query_file, backups)

    # Check if expected equals new_queries
    expected_queries_fn = os.path.join(rundir, '.autoMLSA', 'expected_queries.json')
    if os.path.exists(expected_queries_fn):
        logger.debug('Reading query file {}'.format(expected_queries_fn))
        with open(expected_queries_fn, 'r') as eqf:
            expected_queries: List[str] = json.load(eqf)
    else:
        expected_queries = []

    updated: bool = set(expected_queries) != set(new_queries)
    if updated:
        if all(query in new_queries for query in expected_queries):
            logger.debug('Found new query sequences')
            remove_intermediates(runid, ['query'], protect)
        else:
            logger.debug('Query sequences have been removed')
            remove_intermediates(runid, ['query', 'genome'], protect)
    else:
        logger.debug('Found no new query sequences')
    json_writer(expected_queries_fn, new_queries)

    if not new_queries:
        msg = 'No query sequences found. Check your query file and try again.'
        logger.error(msg)
        end_program(-1)

    open(os.path.join('.autoMLSA', 'checkpoint', 'get_queries'), 'w').close()

    return new_queries



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import subprocess
from .helper_functions import (
    checkpoint_reached,
    checkpoint_tracker,
    end_program,
)
from typing import List


def generate_nexus(runid: str, aligned: List[str], checkpoint: bool) -> str:
    """
    Generate nexus file containing partitions per aligned gene

    input  - list of aligned FASTA filenames
    return - path to nexus file
    """
    logger = logging.getLogger(__name__)
    # Example nexus file
    # #nexus
    # begin sets;
    #         charset part1 = rpoB.all.aln: *;
    #         charset part2 = ppsA.all.aln: *;
    #         charset part3 = gyrB.all.aln: *;
    #         charset part4 = dnaE.all.aln: *;
    #         charset part5 = pyrC.all.aln: *;
    #         charset part6 = aroE.all.aln: *;
    #         charset part7 = acsA.all.aln: *;
    #         charset part8 = recA.all.aln: *;
    # end;

    nexus: str = '{}.nex'.format(runid)
    if os.path.exists(nexus):
        msg = 'Nexus file {} already found, skipping nexus file generation.'
        logger.info(msg.format(nexus))
    else:
        with open(nexus, 'w') as nex:
            nex.write('#nexus\n')
            nex.write('begin sets;\n')
            for fn in aligned:
                base = os.path.basename(os.path.splitext(fn)[0])
                nex.write('\tcharset {} = {}: *;\n'.format(base, fn))
            nex.write('end;\n')
    if checkpoint:
        checkpoint_reached('after generating nexus file {}'.format(nexus))

    checkpoint_tracker('generate_nexus')
    return nexus


def run_iqtree(
    threads: int, iqtree: str, nexus: str, outgroup: str, opts: List[str]
) -> str:
    """
    Runs external iqtree command to generate phylogeny

    input  - nexus file
    return - path to output file
    """
    logger = logging.getLogger(__name__)
    out_tree: str = '{}.treefile'.format(nexus)
    logger.info('Generating phylogenetic tree ({}).'.format(out_tree))
    # cmd: List[str] = [
    #     iqtree,
    #     '-p', nexus,
    #     '-m', 'MFP',
    #     '-B', '1000',
    #     '-alrt', '1000',
    #     '--msub', 'nuclear',
    #     '--merge', 'rclusterf',
    #     '-nt', str(threads)]
    cmd: List[str] = [iqtree, '-p', nexus, '-nt', str(threads)]
    cmd.extend(opts)
    if outgroup:
        cmd.extend(['-o', outgroup])
    if os.path.exists(out_tree):
        logger.info('Treefile {} already found. Skipping iqtree.'.format(out_tree))
    else:
        logger.info('Running: {}'.format(' '.join(cmd)))
        subprocess.run(cmd)
        if not os.path.exists(out_tree):
            msg = 'iqtree2 seems to have failed.'
            logger.critical(msg)
            msg = (
                'Check the log files for error messages to see if they can '
                'be resolved.'
            )
            logger.critical(msg)
            end_program(73)

    checkpoint_tracker('run_iqtree')

    return out_tree



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function
import time
import os
import logging
import argparse
import pandas as pd  # type: ignore
import shlex
from typing import List, Dict, Any
from .parse_args import run_argparse
from .validate_requirements import validate_requirements
from .helper_functions import exit_successfully
from .blast_functions import (
    read_blast_results,
    print_blast_summary,
    print_fasta_files,
    generate_blast_list,
)
from .configuration import read_config, validate_arguments, get_labels
from .formatting import convert_fasta, get_queries
from .mafft import run_mafft
from .phylogeny import generate_nexus, run_iqtree
from signal import signal, SIGPIPE, SIGINT, SIG_DFL
from .__version__ import __version__

signal(SIGPIPE, SIG_DFL)
signal(SIGINT, SIG_DFL)


def main() -> None:
    # ARGPARSE SECTION --------------------------------------------------------
    args: argparse.Namespace = run_argparse()
    logger = logging.getLogger(__name__)
    logger.info('Welcome to autoMLSA.py version {}'.format(__version__))
    time.sleep(1)

    # CONFIGURATION SECTION ---------------------------------------------------
    exes: Dict[str, str] = validate_requirements(args.external)

    logger.info('Reconciling configuration settings.')
    config: Dict[str, Any] = {}
    if os.path.exists(args.configfile):
        config = read_config(args.configfile)
        if args.config is not None:
            msg = 'Config file specified {} is ignored as one is found in {}.'
            logger.warning(msg.format(args.config, args.rundir))
    elif args.config:
        config = read_config(args.config)
    args = validate_arguments(args, config, args.checkpoint == 'validate')

    # FORMATTING SECTION ------------------------------------------------------
    logger.info('Converting genome FASTA files for BLAST if necessary.')
    labels: List[str] = get_labels(args.rundir, args.fasta)
    newfastas: List[str] = convert_fasta(
        args.rundir, args.fasta, labels, exes['makeblastdb'], args.runid, args.protect
    )

    logger.info('Extracting query FASTA files if necessary.')
    queries: List[str] = get_queries(
        args.runid, args.rundir, args.dups, args.query, args.protect
    )

    # BLAST SECTION -----------------------------------------------------------
    logger.info('Generating list of BLAST searches and outputs.')
    blastout: List[str] = generate_blast_list(
        args.rundir,
        exes[args.program],
        queries,
        newfastas,
        args.evalue,
        args.threads,
        args.checkpoint == 'preblast',
    )

    # BLAST output results, summary, and files --------------------------------
    blastres: pd.DataFrame = read_blast_results(
        blastout, args.coverage, args.identity, args.threads
    )
    blastfilt: pd.DataFrame = print_blast_summary(
        args.runid,
        blastres,
        labels,
        args.allow_missing,
        args.missing_check,
        args.checkpoint == 'filtering',
        args.protect,
    )
    unaligned: List[str] = print_fasta_files(blastfilt, labels)

    # ALIGNMENT SECTION -------------------------------------------------------
    aligned: List[str] = run_mafft(
        args.threads, exes['mafft'], unaligned, args.checkpoint, shlex.split(args.mafft)
    )

    # PHYLOGENY SECTION -------------------------------------------------------
    nexus: str = generate_nexus(args.runid, aligned, args.checkpoint == 'nexus')
    treefile: str = run_iqtree(
        args.threads, exes['iqtree'], nexus, args.outgroup, shlex.split(args.iqtree)
    )
    exit_successfully(args.rundir, treefile)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import argparse
import logging
import pprint
import sys
from rich.logging import RichHandler
from rich.console import Console
from .helper_functions import check_if_fasta
from .validate_requirements import (
    check_program,
    validate_requirements,
    install_blast,
    install_mafft,
    install_iqtree,
)
from .__version__ import __version__


class InstallDeps(argparse.Action):
    def __init__(self, nargs='?', **kw):
        super().__init__(nargs=nargs, **kw)

    def __call__(self, parser, namespace, values, option_string=None):
        if values:
            external = values
        else:
            external = os.path.join('~', '.local', 'external')
        if '~' in external:
            external = os.path.expanduser(external)
        external = os.path.abspath(external)
        if not os.path.basename(external) == 'external':
            external = os.path.join(external, 'external')

        init_logger(debug=False, quiet=False, rundir='', runid='')
        logger = logging.getLogger(__name__)
        logger.info('Checking for dependencies and installing.')
        install = False
        if check_program('tblastn', 'tblastn', '-version'):
            logger.info('tblastn already found, skipping install.')
        else:
            install_blast(external)
            install = True
        if check_program('mafft', 'mafft', '--version'):
            logger.info('mafft already found, skipping install.')
        else:
            install_mafft(external)
            install = True
        if check_program('iqtree2', 'iqtree', '--version'):
            logger.info('iqtree2 already found, skipping install.')
        else:
            install_iqtree(external)
            install = True
        validate_requirements(external)
        if install:
            msg = (
                'Please provide given path {} as the --external flag to '
                'use the installed programs.'
            )
            logger.info(msg.format(external))
            msg = '~/.local/external will be found automatically.'
            logger.info(msg)
        parser.exit()


def config_rundir(runid: str) -> str:
    """Identifies existing (or not) run directory from input runid

    input  - runid as given as input
    output - full path to the run directory
    """
    rundir: str = ''
    if os.path.exists('../{}'.format(runid)):
        rundir = os.path.abspath('../{}'.format(runid))
    else:
        rundir = os.path.abspath('./{}'.format(runid))
        if not os.path.exists(rundir):
            try:
                os.mkdir(rundir)
                os.mkdir(os.path.join(rundir, '.autoMLSA'))
                os.mkdir(os.path.join(rundir, '.autoMLSA', 'backups'))
                os.mkdir(os.path.join(rundir, '.autoMLSA', 'checkpoint'))
            except OSError as ose:
                msg = 'Unable to generate rundir "{}" : {}'
                sys.stderr.write(msg.format(rundir, ose) + '\n')
                sys.stderr.write('Check your path and try again.\n')
                exit(71)
    os.chdir(rundir)
    return rundir


def extant_file(x: str) -> str:
    """
    'Type' for argparse - checks that file exists
    input  - path to file
    return - absolute path to file
    """
    if not os.path.exists(x):
        raise argparse.ArgumentTypeError("{} does not exist".format(x))
    return os.path.abspath(x)


def get_fasta_files(x: str) -> str:
    """
    'Type' for argparse - checks if file is FASTA format
    input  - path to file
    return - absolute path to FASTA file
    """
    if not os.path.exists(x):
        raise argparse.ArgumentTypeError("{0} does not exist".format(x))
    elif check_if_fasta(x):
        return os.path.abspath(x)
    else:
        raise argparse.ArgumentTypeError("{0} is not a FASTA file".format(x))


def init_logger(debug: bool, quiet: bool, rundir: str, runid: str) -> None:
    """Sets up logging system to file and stderr

    input  - parsed arguments
    return - None; use logger = logging.getLogger(__name__)
    """

    # must set logging level to DEBUG to print DEBUG to file
    logger = logging.getLogger()
    logger.setLevel(logging.DEBUG)

    stderr_handler = RichHandler(rich_tracebacks=True)
    if debug:
        stderr_handler.setLevel(logging.DEBUG)
    elif quiet:
        stderr_handler.setLevel(logging.WARNING)
    else:
        stderr_handler.setLevel(logging.INFO)
    logger.addHandler(stderr_handler)

    # Print to file as well, if provided.
    if rundir and runid:
        logfile = os.path.join(rundir, '{}.log'.format(runid))
        file_handler = RichHandler(console=Console(file=open(logfile, 'a'), width=119))
        logger.addHandler(file_handler)


def run_argparse() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description='This is a rewrite of autoMLSA.pl. Generates automated '
        'multi-locus sequence analyses.',
        epilog='Quick usage: automlsa2 --dir genomedir --query queries.fas '
        '-t # -- runID',
    )
    parser.add_argument('runid', help='Name of the run directory.', type=str)
    parser.add_argument(
        '--query', help='Path to file with input seq(s).', type=extant_file, nargs='+'
    )
    parser.add_argument(
        '--files',
        help='Path to the target genome FASTA files.',
        type=get_fasta_files,
        nargs='+',
    )
    parser.add_argument(
        '--dir',
        help='Path to the target genome directory with FASTA files.',
        type=extant_file,
        nargs='+',
    )
    parser.add_argument(
        '-e', '--evalue', help='E-value cutoff for BLAST searches. [1e-5]', type=float
    )
    parser.add_argument(
        '-c', '--coverage', help='Sets the coverage cut-off threshold. [50]', type=int
    )
    parser.add_argument(
        '-i', '--identity', help='Sets the identity cut-off threshold. [30]', type=int
    )
    parser.add_argument(
        '-p',
        '--program',
        help='Which BLAST program to run. [tblastn]',
        choices=['blastn', 'tblastn'],
        type=str,
    )
    parser.add_argument(
        '--config', help='Path to configuration json file to copy.', type=extant_file
    )
    parser.add_argument(
        '--missing_check',
        help='Use this to confirm that settings have been '
        'checked when genes are missing.',
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '-t', '--threads', help='Number of threads to use. [1]', type=int
    )
    parser.add_argument(
        '--dups',
        help='Allow for duplicate query names for more sequence '
        'coverage across disparate organisms.',
        action='store_true',
        default=False,
    )
    parser.add_argument(
        '--allow_missing', help='Allow for N missing genes per genome. [0]', type=int
    )
    parser.add_argument(
        '--outgroup', help='Name of outgroup file or strain to root on.', type=str
    )
    parser.add_argument(
        '--protect',
        help='Save files from getting overwritten. By default, '
        'as input files update, older alignments and trees are deleted.',
        action='store_true',
    )
    parser.add_argument(
        '--checkpoint',
        help='Name of stage to stop computing on. [none]',
        type=str,
        choices=[
            'validate',
            'preblast',
            'filtering',
            'prealign',
            'postalign',
            'nexus',
            'none',
        ],
    )
    parser.add_argument(
        '--install_deps',
        help='Install dependencies into given directory. ' '[~/.local/external]',
        action=InstallDeps,
    )
    parser.add_argument(
        '--external',
        help='Path to installed external programs. ' '[~/.local/external]',
        type=extant_file,
    )
    parser.add_argument(
        '--mafft', help='mafft settings [--localpair --maxiterate 1000]', type=str
    )
    parser.add_argument(
        '--iqtree',
        help='iqtree2 settings '
        '[-m MFP -B 1000 -alrt 1000 --msub nuclear --merge rclusterf]',
        type=str,
    )
    parser.add_argument(
        '--debug',
        help='Turn on debugging messages.',
        action='store_true',
        default=False,
    )
    parser.add_argument('--version', action='version', version=__version__)
    parser.add_argument(
        '--quiet',
        help='Turn off progress messages.',
        action='store_true',
        default=False,
    )
    args = parser.parse_args()
    args.rundir = config_rundir(args.runid)
    args.configfile = os.path.join(args.rundir, 'config.json')
    init_logger(args.debug, args.quiet, args.rundir, args.runid)
    logger = logging.getLogger(__name__)
    logger.debug('Started autoMLSA.py for run: {}'.format(args.runid))
    logger.debug(pprint.pformat(vars(args)))
    if not args.external:
        args.external = ''
    return args



# -*- coding: utf-8 -*-
from setuptools import find_packages, setup  # type: ignore

with open('automlsa2/__version__.py', 'r') as f:
    for line in f:
        if line.startswith('__version__'):
            version = line.strip().split('=')[1].strip(' \'"')
            break
    else:
        version = '0.0.1'

with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()

REQUIRES = ['biopython', 'pandas>=1.0.0', 'numpy', 'rich']

kwargs = {
    'name': 'automlsa2',
    'version': version,
    'description': 'Automated Multi-Locus Sequence Analysis tool',
    'long_description_content_type': 'text/x-rst',
    'long_description': readme,
    'author': 'Edward Davis',
    'author_email': 'ed@cgrb.oregonstate.edu',
    'maintainer': 'Edward Davis',
    'maintainer_email': 'ed@cgrb.oregonstate.edu',
    'url': 'https://github.com/davised/automlsa2',
    'license': 'Custom',
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: Free for non-commercial use',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
    ],
    'install_requires': REQUIRES,
    'tests_require': ['coverage', 'pytest'],
    'packages': find_packages(exclude=('tests', 'tests.*')),
}

setup(
    **kwargs, entry_points={"console_scripts": ['automlsa2 = automlsa2.__main__:main']}
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from __future__ import print_function
import subprocess
import numpy as np  # type: ignore
import pandas as pd  # type: ignore
import json
import os
import logging
import multiprocessing
import shlex
import csv
import sys
from multiprocessing import Pool
from typing import List, Dict, Any, DefaultDict
from .helper_functions import (
    json_writer,
    checkpoint_reached,
    checkpoint_tracker,
    remove_intermediates,
    end_program,
    SUFFIXES,
)
from collections import defaultdict
from rich.progress import track
from signal import signal, SIGPIPE, SIGINT, SIG_DFL

signal(SIGPIPE, SIG_DFL)
signal(SIGINT, SIG_DFL)

SINGLE_COPY_ESTIMATE = 0.90
PERCENT_CUTOFF = 50.0


def make_blast_database(makeblastdb: str, fasta: str) -> None:
    """Generates blast DB if necessary per fasta

    input  - fasta file
    return - null
    """
    exist: List[bool] = [os.path.exists(fasta + x) for x in SUFFIXES]
    if not all(exist):
        with open(os.path.join('fasta', 'makeblastdb.log'), 'a') as textlog:
            subprocess.run(
                [makeblastdb, '-dbtype', 'nucl', '-in', fasta],
                stdout=textlog,
                stderr=subprocess.STDOUT,
            )


def reader(fn: str) -> List[List[str]]:
    """
    Read tsv file into list
    input  - filename to read
    return - list of results
    """
    dat: List[List[str]] = []
    with open(fn, 'r') as tsv:
        tsvreader = csv.reader(
            [row for row in tsv if not row.startswith('#')], delimiter='\t'
        )
        for row in tsvreader:
            dat.append(row)
    return dat


def read_blast_results(
    blastfiles: List[str], coverage: int, identity: int, threads: int
) -> pd.DataFrame:
    """
    Parses set of BLAST results. Expects list of files.

    input  - list of BLAST output filenames.
    return - pandas data frame of BLAST output
    """
    logger: logging.Logger = logging.getLogger(__name__)
    headers: List[str] = [
        'qseqid',
        'sseqid',
        'saccver',
        'pident',
        'qlen',
        'length',
        'bitscore',
        'qcovhsp',
        'stitle',
        'sseq',
    ]

    logger.info('Reading BLAST results.')
    dtypes: Dict[str, Any] = {
        'qseqid': 'category',
        'sseqid': 'category',
        'saccver': 'string',
        'pident': np.float,
        'qlen': np.int,
        'length': np.int,
        'bitscore': np.float,
        'qcovhsp': np.int,
        'stitle': 'string',
        'sseq': 'string',
    }

    results_fn: str = os.path.join('.autoMLSA', 'blast_results.tsv')
    if os.path.exists(results_fn):
        logger.debug('Reading from existing BLAST results.')
        results: pd.DataFrame = pd.read_csv(results_fn, dtype=dtypes, sep='\t')
    else:
        # results = pd.DataFrame(columns=headers)
        blastrows: List[List[str]] = []
        nfiles = len(blastfiles)
        p: multiprocessing.pool.Pool = Pool(threads)
        with p:
            for dat in track(
                p.imap_unordered(reader, blastfiles),
                'Reading Blast...'.rjust(19, ' '),
                nfiles,
            ):
                for row in dat:
                    blastrows.append(row)
        results = pd.DataFrame(data=blastrows, columns=headers)
        results = results.astype(dtypes)
        results.query('(pident >= @identity) & (qcovhsp >= @coverage)', inplace=True)
        results.to_csv(results_fn, sep='\t')
    checkpoint_tracker('read_blast_results')
    return results


def print_blast_summary(
    runid: str,
    blastout: pd.DataFrame,
    labels: List[str],
    nallowed: int,
    missing_check: bool,
    checkpoint: bool,
    protect: bool,
) -> pd.DataFrame:
    """
    Generates summary of BLAST results.

    input  - pandas data frame with BLAST results
    return - pandas data frame with BLAST results to keep
    """
    logger: logging.Logger = logging.getLogger(__name__)
    logger.info('Summarizing and filtering BLAST hits.')
    summary: DefaultDict[str, Dict[str, Any]] = defaultdict(dict)
    missing_count: DefaultDict[int, List[str]] = defaultdict(list)
    genome_missing_dict: Dict = {}
    keeps: List[str] = []
    keepsidx: List[str] = []
    genome_map: Dict[str, int] = {k: i for i, k in enumerate(labels)}

    grouped: pd.DataFrame = blastout.groupby(['sseqid', 'qseqid'])
    presence_matrix: pd.DataFrame = grouped.size().unstack('qseqid', fill_value=0)
    label_indexes = list(map(int, presence_matrix.index.values))
    indexes: List[int] = presence_matrix.index.tolist()
    presence_matrix.index = [labels[i] for i in label_indexes]
    presence_matrix.to_csv('presence_matrix.tsv', sep='\t')
    summary['queries']['names'] = presence_matrix.columns.tolist()
    summary['queries']['count'] = len(presence_matrix.columns)
    summary['genomes']['names'] = presence_matrix.index.tolist()  # type: ignore
    summary['genomes']['indexes'] = indexes
    summary['genomes']['count'] = len(presence_matrix.index)

    # Get zero counts for filtering
    # Queries
    removal_candidates: Dict[str, str] = {}
    summary['queries']['missing'] = defaultdict(dict)
    zero_counts_query: pd.DataFrame = presence_matrix.apply(
        lambda x: x[x == 0].index.tolist(), axis=0
    )
    zero_counts_query_dict: Dict[str, List[str]] = zero_counts_query.to_dict()
    for query, genomes in zero_counts_query_dict.items():
        ngenomes = len(genomes)
        percent = ngenomes / summary['genomes']['count'] * 100
        percentf = '{:.2f}'.format(percent)
        if percent > PERCENT_CUTOFF:
            removal_candidates[query] = percentf
        summary['queries']['missing'][query]['genomes'] = genomes
        summary['queries']['missing'][query]['count'] = ngenomes
        summary['queries']['missing'][query]['percent'] = percentf
    summary['queries']['missing'] = dict(summary['queries']['missing'])

    # Genomes
    summary['genomes']['missing'] = defaultdict(dict)
    zero_counts_genome: pd.DataFrame = presence_matrix.apply(
        lambda x: x[x == 0].index.tolist(), axis=1
    )
    zero_counts_genome_dict: Dict[str, List[str]] = zero_counts_genome.to_dict()
    for genome, queries in zero_counts_genome_dict.items():
        nmissing: int = len(queries)
        summary['genomes']['missing'][genome]['queries'] = queries
        summary['genomes']['missing'][genome]['count'] = nmissing
        summary['genomes']['missing'][genome]['percent'] = '{:.2f}'.format(
            nmissing / summary['queries']['count'] * 100
        )
        missing_count[nmissing].append(genome)
        if nmissing == 0:
            keeps.append(genome)
        else:
            genome_missing_dict[genome] = queries
    summary['genomes']['missing'] = dict(summary['genomes']['missing'])

    for genome in genome_missing_dict:
        nmissing = len(genome_missing_dict[genome])
        if nmissing > nallowed:
            msg = 'Genome {} is going to be removed due to missing queries.'
            logger.warning(msg.format(genome))
            msg = 'Increase --allow_missing to {} from {} to keep this genome.'
            logger.warning(msg.format(nmissing, nallowed))
        else:
            msg = 'Keeping genome {} missing {} queries due to --allow_missing'
            logger.debug(msg.format(genome, nmissing))
            keeps.append(genome)
    msg = 'Keeping these genomes ({}):\n         - {}'
    logger.info(msg.format(len(keeps), '\n         - '.join(keeps)))
    keepsidx = [str(genome_map[x]) for x in keeps]
    json_writer(os.path.join('.autoMLSA', 'keepsidx.json'), keepsidx)
    blastout_keeps: pd.DataFrame = (
        blastout.query('sseqid in @keepsidx')
        .sort_values(['bitscore', 'qcovhsp'], ascending=False)
        .drop_duplicates(['qseqid', 'sseqid'])
        .sort_index()
    )

    blastout_keeps.to_csv(os.path.join('.autoMLSA', 'blast_filtered.tsv'), sep='\t')

    # Estimate single copy
    # Currently UNUSED #
    single_copy: pd.DefaultDict = presence_matrix.apply(
        lambda x: ((x.values == 1).sum() / len(presence_matrix)) > SINGLE_COPY_ESTIMATE,
        axis=0,
    )
    json_writer(os.path.join('.autoMLSA', 'single_copy.json'), single_copy.to_dict())
    json_writer('missing_counts.json', missing_count)
    json_writer('blast_summary.json', summary)
    json_writer('missing_by_genome.json', genome_missing_dict)
    if len(keeps) < summary['genomes']['count']:
        logger.warning(
            'Some genomes ({}) are missing one or more genes'.format(
                len(genome_missing_dict)
            )
        )
        logger.warning(
            'Check out the presence_matrix.tsv, and '
            'missing_by_genome.json files to review.'
        )
        if removal_candidates:
            logger.warning('Consider removing these genes from the analysis:')
            for query, percent in removal_candidates.items():
                sys.stderr.write(f'         - {query} -> missing {percent}%\n')
        if not missing_check:
            logger.warning('Use --missing_check to continue.')
            end_program(1)
    expected_filt_fn: str = os.path.join('.autoMLSA', 'expected_filt.json')
    if os.path.exists(expected_filt_fn):
        with open(expected_filt_fn, 'r') as eqf:
            expected_filt: List[str] = json.load(eqf)
    else:
        expected_filt = []

    updated = set(expected_filt) != set(keeps)
    if updated:
        logger.debug('Found new filtered sequences')
        logger.debug('Removing downstream files, if present')
        remove_intermediates(runid, ['genome'], protect)
    if not updated:
        logger.debug('Found no new filtered sequences')
    json_writer(expected_filt_fn, keeps)

    if checkpoint:
        checkpoint_reached('after BLAST result filtering')

    checkpoint_tracker('print_blast_summary')

    return blastout_keeps


def print_fasta_files(blastout: pd.DataFrame, labels: List[str]) -> List[str]:
    """
    Writes unaligned FASTA files as output from BLAST results

    input  - pandas DataFrame with blast output, list of labels
    return - List of unaligned FASTA files
    """
    logger: logging.Logger = logging.getLogger(__name__)
    fastdir: str = 'unaligned'
    labels = [os.path.splitext(x)[0] for x in labels]
    unaligned: List[str] = []
    if not os.path.exists(fastdir):
        os.mkdir(fastdir)
    msg = 'Writing unaligned FASTA sequences, if necessary.'
    logger.info(msg)
    for name, group in track(blastout.groupby('qseqid')):
        fasta: str = os.path.join(fastdir, '{}.fas'.format(name))
        unaligned.append(fasta)
        if os.path.exists(fasta):
            logger.debug('Unaligned file {} already found, skipping.'.format(fasta))
        else:
            if os.path.exists(fasta):
                text = 'Overwriting'
            else:
                text = 'Writing'
            logger.debug('{} {} FASTA file'.format(text, fasta))
            with open(fasta, 'w') as fh:
                for row in group.itertuples():
                    fh.write('>{}\n'.format(labels[int(row.sseqid)]))
                    fh.write('{}\n'.format(row.sseq.replace('-', '')))

    checkpoint_tracker('print_fasta_files')

    return unaligned


def generate_blast_list(
    rundir: str,
    exe: str,
    queries: List[str],
    targets: List[str],
    evalue: float,
    threads: int,
    checkpoint: bool,
) -> List[str]:
    """
    Generates list of BLAST searches that need to be run

    input  - blast type, exes, queries, and targets
    output - list of blast output names
    """
    logger = logging.getLogger(__name__)
    cmds: List[List[str]] = []
    # blastout = defaultdict(list)
    blastout: List[str] = []
    blastdir: str = os.path.join(rundir, 'blast')
    outfmt: str = (
        '7 qseqid sseqid saccver pident qlen length bitscore ' 'qcovhsp stitle sseq'
    )
    base_cmd: List[str] = [exe, '-evalue', str(evalue), '-outfmt', outfmt]
    cmd: List[str] = []
    if not os.path.exists(blastdir):
        os.mkdir(blastdir)
    for target in targets:
        target_base: str = os.path.basename(target)
        for query in queries:
            query_base: str = os.path.splitext(os.path.basename(query))[0]
            outname: str = '{}_vs_{}.tab'.format(query_base, target_base)
            outpath: str = os.path.join(blastdir, outname)
            blastout.append(os.path.join(outpath))
            cmd = base_cmd + ['-db', target, '-query', query, '-out', outpath]
            if not os.path.exists(outpath) or os.path.getsize(outpath) == 0:
                cmds.append(cmd)

    if cmds:
        if checkpoint:
            blastfile: str = os.path.join(rundir, 'blastcmds.txt')
            with open(blastfile, 'w') as fh:
                for cmd in cmds:
                    fh.write(' '.join([shlex.quote(x) for x in cmd]) + '\n')
            msg = 'BLAST commands written to {}. Exiting.'
            logger.info(msg.format(blastfile))
            checkpoint_reached('prior to BLAST searches')
        else:
            msg = 'Running {} BLAST searches using {} CPUs.'
            ncmds: int = len(cmds)
            logger.info(msg.format(ncmds, threads))
            p: multiprocessing.pool.Pool = Pool(threads)
            with p:
                for _ in track(
                    p.imap_unordered(subprocess.run, cmds),
                    'Blast Search...'.rjust(19, ' '),
                    ncmds,
                ):
                    pass
    else:
        logger.info('No BLAST searches remaining. Moving to parse.')

    open(os.path.join('.autoMLSA', 'checkpoint', 'generate_blast_list'), 'w').close()
    return blastout



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
import logging
import shlex
import subprocess
from .helper_functions import checkpoint_reached, checkpoint_tracker
from rich.progress import track
from typing import List


def run_mafft(
    threads: int,
    mafft: str,
    unaligned: List[str],
    checkpoint: bool,
    opts: List[str],
) -> List[str]:
    """
    input  - unaligned fasta files per query
    return - list of aligned files per query
    """
    logger = logging.getLogger(__name__)
    base_cmd: List[str] = [mafft, '--thread', str(threads)]
    base_cmd.extend(opts)
    aligneddir: str = 'aligned'
    aligned: List[str] = []
    cmdstrs: List[str] = []
    logger.info('Aligning FASTA sequences, if necessary.')
    if not os.path.exists(aligneddir):
        os.mkdir(aligneddir)
    for unalign in track(unaligned, 'Running mafft...'.rjust(19, ' ')):
        outname: str = '{}/{}.aln'.format(
            aligneddir, os.path.basename(os.path.splitext(unalign)[0])
        )
        logname: str = '{}.log'.format(outname)
        logger.debug('Aligning and writing to {}'.format(outname))
        aligned.append(outname)
        if os.path.exists(outname):
            msg = 'Aligned file {} already found, skipping.'
            logger.debug(msg.format(outname))
        else:
            cmd: List[str] = base_cmd + [unalign]
            cmdstr = ' '.join([shlex.quote(x) for x in cmd]) + ' > {}'.format(
                shlex.quote(outname)
            )
            logger.debug(cmdstr)
            if checkpoint == 'prealign':
                cmdstrs.append(cmdstr)
            else:
                with open(outname, 'w') as fh, open(logname, 'w') as logfh:
                    subprocess.run(cmd, stdout=fh, stderr=logfh, text=True)
    if cmdstrs:
        with open('mafftcmds.txt', 'w') as fh:
            fh.write('\n'.join(cmdstrs))
        logger.info('MAFFT alignment commands written to mafftcmds.txt.')
        logger.info('Run these commands and resubmit to continue.')
        checkpoint_reached('prior to MAFFT alignments')
    elif checkpoint == 'postalign':
        checkpoint_reached('after MAFFT alignments')

    checkpoint_tracker('run_mafft')

    return aligned
